/*
 * Copyright (C) 2022 Jonathan Certes
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>
#include <vpi_user.h>

/**
 * \brief
 *  Implementation of the increment verilog task
 */
static int32_t myCFunc_increment(
  char* theUserData //!< Pointer to the user data
) {
  int32_t theLocalValue;  // to manipulate an integer locally
  int32_t i;              // to add more than 1

  vpiHandle theTaskHandle, theArgsListHandle, theArgHandle;
  struct t_vpi_value theArgStruct;

  if ( strlen(theUserData) > 0 ) {
    vpi_printf("The user data: %s\n", theUserData);
  }

  // obtain a handle to the argument list:
  theTaskHandle     = vpi_handle(vpiSysTfCall, NULL);
  theArgsListHandle = vpi_iterate(vpiArgument, theTaskHandle);

  i = 0;
  for (
    theArgHandle  = vpi_scan(theArgsListHandle);
    theArgHandle != NULL;
    theArgHandle  = vpi_scan(theArgsListHandle)
  ) {
    // grab the value of the argument as an integer:
    theArgStruct.format = vpiIntVal;
    vpi_get_value(theArgHandle, &theArgStruct);
    theLocalValue = theArgStruct.value.integer;

    vpi_printf("VPI routine received %d\n", theLocalValue);

    // increment the value and put it back as argument:
    theArgStruct.value.integer = theLocalValue + 1 + i;
    vpi_put_value(theArgHandle, &theArgStruct, NULL, vpiNoDelay);
    i++;
  }

  // cleanup:
  vpi_free_object(theTaskHandle);
  // No need to free theArgsListHandle and theArgHandle: the iterator objects
  // shall automatically be freed when vpi_scan() returns NULL.
  return 0;
}

/******************************************************************************/

/**
 * \brief
 *  Adds a function to the registered verilog tasks.
 */
void register_increment(
  void
) {
  s_vpi_systf_data theTaskData;
  theTaskData.type      = vpiSysTask;
  theTaskData.tfname    = "$increment";
  theTaskData.calltf    = myCFunc_increment;
  theTaskData.compiletf = NULL;
  theTaskData.user_data = "My user data";
  vpi_register_systf(&theTaskData);
}

/******************************************************************************/

/**
 * \brief
 *  Places the register function in the externally visible
 *  vlog_startup_routines[] array.
 *
 * \details
 *  Contains a zero-terminated list of functions that have to be called at
 *  startup.
 */
void (*vlog_startup_routines[])( void ) = {
  register_increment,
  0
};

/******************************************************************************/

